from pydantic import BaseModel as _BaseModel


class BaseModel(_BaseModel):

    class Config:
        @staticmethod
        def alias_generator(snake_case_string):
            """to camelCase"""
            words = snake_case_string.split('_')
            return words[0] + ''.join(word.title() for word in words[1:])

    def dict(self, *args, **kwargs):
        if "by_alias" in kwargs:
            del kwargs["by_alias"]
        return super().dict(*args, by_alias=True, **kwargs)

    def json(self, *args, **kwargs):
        if "by_alias" in kwargs:
            del kwargs["by_alias"]
        return super().json(*args, by_alias=True, **kwargs)



from typing import Callable, Dict, List
import abc
import copy
import operator


empty = object()


def method_proxy(func):
    def inner(self, *args):
        if self._wrapped is empty:
            self._setup()
        return func(self._wrapped, *args)
    return inner


class Lazy:
    """
    A wrapper for another class that can be used to delay instantiation of the
    wrapped class. The wrapper can provide properties that are already known
    (without instantiating the wrapped class).
    """

    # Avoid infinite recursion when tracing __init__
    _wrapped = None

    def __init__(self, getter: Callable, known_properties: Dict = None, known_non_properties: List = None):
        # Note: if a subclass overrides __init__(), it will likely need to
        # override __copy__() and __deepcopy__() as well.
        self.__dict__['_get_wrapped'] = getter
        self._wrapped = empty
        self.__dict__["_known_properties"] = known_properties if known_properties else {}
        self.__dict__["_known_non_properties"] = known_non_properties if known_non_properties else []

    def __getattr__(self, name):
        if self._wrapped is empty and name in self._known_properties:
            return self._known_properties[name]
        else:
            if self._wrapped is empty:
                if name in self._known_non_properties:
                    raise AttributeError()
                self._setup()
            return getattr(self._wrapped, name)

    def __setattr__(self, name, value):
        if name == "_wrapped":
            # Assign to __dict__ to avoid infinite __setattr__ loops
            self.__dict__["_wrapped"] = value
        else:
            if self._wrapped is empty:
                self._setup()
            setattr(self._wrapped, name, value)

    def __delattr__(self, name):
        if name == "_wrapped":
            raise TypeError("Can't delete _wrapped.")
        if self._wrapped is empty:
            if name in self._known_non_properties:
                raise AttributeError()
            self._setup()
        delattr(self._wrapped, name)

    @abc.abstractmethod
    def _setup(self):
        self._wrapped = self._get_wrapped()

    def __reduce__(self):
        # Because we have messed with __class__ below, we confuse pickle as to what
        # class we are pickling. We're going to have to initialize the wrapped
        # object to successfully pickle it, so we might as well just pickle the
        # wrapped object since they're supposed to act the same way.
        #
        # Unfortunately, if we try to simply act like the wrapped object, the ruse
        # will break down when pickle gets our id(). Thus we end up with pickle
        # thinking, in effect, that we are a distinct object from the wrapped
        # object, but with the same __dict__. This can cause problems (see #25389).
        #
        # So instead, we define our own __reduce__ method and custom unpickler. We
        # pickle the wrapped object as the unpickler's argument, so that pickle
        # will pickle it normally, and then the unpickler simply returns its
        # argument.
        def unpickle_lazyobject(wrapped):
            return wrapped
        if self._wrapped is empty:
            self._setup()
        return unpickle_lazyobject, (self._wrapped,)

    def __copy__(self):
        if self._wrapped is empty:
            # If uninitialized, copy the wrapper. Use type(self), not
            # self.__class__, because the latter is proxied.
            return type(self)(self._get_object, self._known_properties)
        else:
            # If initialized, return a copy of the wrapped object.
            return copy.copy(self._wrapped)

    def __deepcopy__(self, memo):
        if self._wrapped is empty:
            # We have to use type(self), not self.__class__, because the
            # latter is proxied.
            result = type(self)(self._get_object, self._known_properties)
            memo[id(self)] = result
            return result
        return copy.deepcopy(self._wrapped, memo)

    __bytes__ = method_proxy(bytes)
    __str__ = method_proxy(str)
    __bool__ = method_proxy(bool)

    # Introspection support
    __dir__ = method_proxy(dir)

    # Need to pretend to be the wrapped class, for the sake of objects that
    # care about this (especially in equality tests)
    __class__ = property(method_proxy(operator.attrgetter("__class__")))
    __eq__ = method_proxy(operator.eq)
    __lt__ = method_proxy(operator.lt)
    __gt__ = method_proxy(operator.gt)
    __ne__ = method_proxy(operator.ne)
    __hash__ = method_proxy(hash)

    # List/Tuple/Dictionary methods support
    __getitem__ = method_proxy(operator.getitem)
    __setitem__ = method_proxy(operator.setitem)
    __delitem__ = method_proxy(operator.delitem)
    __iter__ = method_proxy(iter)
    __len__ = method_proxy(len)
    __contains__ = method_proxy(operator.contains)

    def __repr__(self):
        if self._wrapped is empty:
            return f"{type(self).__name__} -> {self._setupfunc}"
        else:
            return repr(self._wrapped)



from typing import Union, Dict
import os
import logging
from aws_lambda_powertools.middleware_factory import lambda_handler_decorator
from aws_lambda_powertools import Logger
from aws_lambda_powertools.utilities.parser import ValidationError
from . import BaseModel


class BadOperationUserError(Exception):
    pass


class EntityDoesNotExistUserError(Exception):
    pass


class AppErrorModel(BaseModel):
    type: str
    message: Union[str, Dict]


class AppErrorResponseModel(BaseModel):
    app_error: AppErrorModel = None


env_service_name = os.environ.get("Service")
if env_service_name is None:
    raise ValueError("'Service' env var not set (service name)")
env_logging_level = os.environ.get("LoggingLevel")
if env_logging_level is None:
    env_logging_level = logging.INFO
else:
    env_logging_level = int(env_logging_level)

logger = Logger(service=env_service_name, level=env_logging_level)


def error(type_: str, message: Union[str, Dict]):
    return AppErrorResponseModel(app_error=AppErrorModel(type=type_, message=message)).dict()


@lambda_handler_decorator
def middleware(handler, event, context):
    handler = logger.inject_lambda_context(log_event=logger.level == logging.DEBUG)(handler)
    try:
        response = handler(event, context)
        logger.debug("Handler Response", extra=response)
        return response
    except ValidationError as e:
        logger.info("EventValidationUserError")
        return error("EventValidationUserError", e.json())
    except BadOperationUserError as e:
        logger.info("BadOperationUserError")
        return error("BadOperationUserError", str(e))
    except EntityDoesNotExistUserError as e:
        logger.info("EntityDoesNotExistUserError")
        return error("EntityDoesNotExistUserError", str(e))
    except:
        logger.exception("InternalError")
        return error("InternalError", "Oh no!")


@lambda_handler_decorator
def middleware_mock(handler, event, context):
    response = handler(event, context)
    return response



from .base_model import BaseModel
from .lazy import Lazy
from .middleware import middleware, BadOperationUserError, EntityDoesNotExistUserError, logger, middleware_mock

try:
    from .communication import http
except ImportError:
    pass



try:
    import requests
    import urllib3.util.retry

    http = requests.Session()

    retry = urllib3.util.retry.Retry(
        total=3,
        backoff_factor=0.2,
        allowed_methods=["HEAD", "GET", "OPTIONS"],
        status_forcelist=[429, 500, 502, 503, 504],
        raise_on_status=True,
        redirect=3,
        raise_on_redirect=True,
    )
    retry.BACKOFF_MAX = 2

    adapter = requests.adapters.HTTPAdapter(max_retries=retry)
    http.mount("https://", adapter)
    http.mount("http://", adapter)

except ImportError:
    pass
